from bolib3 import np

"""
A basic example where the upper and lower level programs are totally unlinked.
minimise    x1^2 + x2^2 + x3^2 + x4^2
subject to  x1 >= +1.0,
            x2 >= -2.0,
            x3 == +3.0,
            x4 == -4.0,
            y = argmin(y1^2 + y2^2 + y3^2 + y4^2),
            y1 >= +1.0,
            y2 >= -2.0,
            y3 == +3.0,
            y4 == -4.0,
            
"""

# Properties
name: str = 'basic_constrained'
category: str = 'miscellaneous'
subcategory: str = ''
datasets: list = []

# Feasible point
x0 = [1.0, 0.0, 3.0, -4.0]
y0 = [1.0, 0.0, 3.0, -4.0]


# Methods
def F(x, y, data=None):
    """
    Upper-level objective function
    """
    return np.sum(np.square(x))


def G(x, y, data=None):
    """
    Upper-level inequality constraints
    """
    return np.array([
        x[0] - 1,
        x[1] + 2
    ])


def H(x, y, data=None):
    """
    Upper-level equality constraints
    """
    return np.array([
        x[2] - 3,
        x[3] + 4
    ])


def f(x, y, data=None):
    """
    Lower-level objective function
    """
    return np.sum(np.square(y))


def g(x, y, data=None):
    """
    Lower-level inequality constraints
    """
    return np.array([
        y[0] - 1,
        y[1] + 2,
    ])


def h(x, y, data=None):
    """
    Lower-level equality constraints
    """
    return np.array([
        y[2] - 3,
        y[3] + 4,
    ])


def read_data(filepath=''):
    """
    If the bilevel program is parameterized by data, this function should
     provide code to read data file and return an appropriate python structure.
    """
    pass



def dimensions(key='', data=None):
    """
    If the argument 'key' is not specified, then:
     - a dictionary mapping variable/function names (str) to the corresponding dimension (int) is returned.
    If the first argument 'key' is specified, then:
     - a single integer representing the dimension of the variable/function with the name {key} is returned.
    """
    n = {
        "x": 4,  # Upper-level variables
        "y": 4,  # Lower-level variables
        "F": 1,  # Upper-level objective functions
        "G": 2,  # Upper-level inequality constraints
        "H": 2,  # Upper-level equality constraints
        "f": 1,  # Lower-level objective functions
        "g": 2,  # Lower-level inequality constraints
        "h": 2,  # Lower-level equality constraints
    }
    if key in n:
        return n[key]
    return n
